from typing import Any, Dict, Type

import pytest

import datasets
from molflux.modelzoo.catalogue import register_model
from molflux.modelzoo.info import ModelInfo
from molflux.modelzoo.load import load_model
from molflux.modelzoo.model import ModelBase, ModelConfig
from molflux.modelzoo.protocols import Model


class MockModelConfig(ModelConfig):
    ...


# Registers a mock model in the catalogue
@register_model(kind="pytest", name="mock_model")
class MockModel(ModelBase[MockModelConfig]):
    @property
    def _config_builder(self) -> Type[MockModelConfig]:
        return MockModelConfig

    def _info(self) -> ModelInfo:
        return ModelInfo(
            model_description="A mock model",
        )

    def _train(self, train_data: datasets.Dataset, **kwargs: Any) -> None:
        self.model = True

    def _predict(self, data: datasets.Dataset, **kwargs: Any) -> Dict[str, Any]:
        return {self.tag: "Spooky Result!"}


@pytest.fixture()
def fixture_mock_model_untrained() -> Model:
    """Returns an untrained mock model."""
    return load_model("mock_model", tag="custom_tag")


@pytest.fixture()
def fixture_mock_model_trained() -> Model:
    """Returns a trained mock model."""
    model = load_model(
        "mock_model",
        tag="custom_tag",
        x_features=["x"],
        y_features=["y"],
    )
    model.train(datasets.Dataset.from_dict({"x": [1], "y": [1]}))
    return model



import inspect
import logging
from abc import ABC, abstractmethod
from dataclasses import asdict
from typing import Any, Dict, Optional, Tuple

import evaluate

import datasets
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)


class HFMetric(evaluate.Metric, ABC):
    """A base class for metrics based on the evaluate.Metric API."""

    def __init__(
        self,
        tag: Optional[str] = None,
        config_name: Optional[str] = None,
        keep_in_memory: bool = False,
        seed: Optional[int] = None,
        experiment_id: Optional[str] = None,
        download_config: Optional[datasets.DownloadConfig] = None,
        **kwargs: Any,
    ) -> None:
        # Initialise the evaluate.Metric
        super().__init__(
            config_name=config_name,
            keep_in_memory=keep_in_memory,
            cache_dir=None,
            num_process=1,
            process_id=0,
            seed=seed,
            experiment_id=experiment_id,
            max_concurrent_cache_files=10000,
            timeout=100,
            **kwargs,
        )

        # Download and prepare resources for the metric
        self.download_and_prepare(download_config=download_config)

        self._tag = tag or self.name

        # Initialise a null state
        self._state: Dict[str, Any] = {}

        # The metric-specific compute signature
        self._signature = inspect.signature(self._score)

    def __bool__(self) -> bool:
        return True

    def __len__(self) -> int:
        length: int = super().__len__()
        return length

    def __str__(self) -> str:
        return (
            f"Metric(\n"
            f'\tname: "{self.name}",\n'
            f'\ttag: "{self.tag}",\n'
            f"\tfeatures: {self.features},\n"
            f"\tsignature: self.compute{self._signature},\n"
            f'\tdescription: """{self.description}""",\n'
            f'\tusage: """{self.inputs_description}"""\n'
            f"\tstate: {self.state!r}\n"
            ")"
        )

    @property
    def metadata(self) -> Dict[str, Any]:
        return asdict(super().info)

    @abstractmethod
    def _info(self) -> evaluate.MetricInfo:
        """Initialises the MetricInfo metadata object for the given metric."""

    @property
    def name(self) -> str:
        metric_name: str = super().name
        return metric_name

    @property
    def tag(self) -> str:
        return self._tag

    @property
    def state(self) -> Dict[str, Any]:
        return self._state

    def add_batch(
        self,
        *,
        predictions: Optional[ArrayLike] = None,
        references: Optional[ArrayLike] = None,
        **kwargs: Any,
    ) -> None:
        """Add a batch of predictions and references for the metric's stack."""
        predictions, references = self._pre_process_inputs(
            predictions=predictions,
            references=references,
        )
        super().add_batch(predictions=predictions, references=references, **kwargs)

    def compute(
        self,
        *,
        predictions: Optional[ArrayLike] = None,
        references: Optional[ArrayLike] = None,
        **kwargs: Any,
    ) -> MetricResult:
        """Computes the metric."""

        # Merge explicit keyword arguments with those stored in config
        if self._state is not None:
            kwargs = {**self._state, **kwargs}

        # Safeguard against invalid kwargs
        if not all(k in self._signature.parameters for k in kwargs):
            unknown_kwargs = [k for k in kwargs if k not in self._signature.parameters]
            raise ValueError(
                f"Unknown compute parameter(s): {unknown_kwargs}\n\n"
                f"Expected signature self.compute{self._signature}",
            )

        predictions, references = self._pre_process_inputs(
            predictions=predictions,
            references=references,
        )

        # wrap huggingface's implementation
        try:
            # This loads the data from the cache into self.data,
            # and then calls ._compute()
            result: Optional[MetricResult] = super().compute(
                predictions=predictions,
                references=references,
                **kwargs,
            )

            # On distributed systems, huggingface would return None
            if result is None:
                raise RuntimeError

        except Exception as exc:
            raise RuntimeError(f"Error computing metric {self.tag!r}") from exc

        return result

    def _compute(
        self,
        *,
        predictions: Optional[ArrayLike] = None,
        references: Optional[ArrayLike] = None,
        **kwargs: Any,
    ) -> MetricResult:
        """The callable invoked by .compute() to compute the metric."""

        # Do some checks common to all metrics
        if predictions is None or references is None:
            raise RuntimeError(
                f"""Could not compute metric {self.tag}: no data found in the stack.""",
            )

        return self._score(
            predictions=predictions,
            references=references,
            **kwargs,
        )

    @abstractmethod
    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        **kwargs: Any,
    ) -> MetricResult:
        """The callable invoked to compute the metric. To be implemented by subclasses."""

    def reset_state(self) -> None:
        """Resets the state."""
        self._state = {}

    def update_state(self, **kwargs: Any) -> None:
        """Pre-configures keyword arguments for compute()."""
        self._state = {**self._state, **kwargs}

    def _pre_process_inputs(
        self,
        predictions: Optional[ArrayLike] = None,
        references: Optional[ArrayLike] = None,
    ) -> Tuple[Optional[ArrayLike], ...]:
        """An optional callable used to process inputs before validation.

        For most use cases, there is no need to overwrite this function.
        Defaults to pass-through.
        """
        return predictions, references


class PredictionIntervalMetric(HFMetric):
    """
    A class for uncertainty metrics.
    Same as HFMetric but with a different signature (eg. for .compute()) so can take in prediction intervals
    """

    def __init__(
        self,
        tag: Optional[str] = None,
        config_name: Optional[str] = None,
        keep_in_memory: bool = False,
        seed: Optional[int] = None,
        experiment_id: Optional[str] = None,
        download_config: Optional[datasets.DownloadConfig] = None,
        **kwargs: Any,
    ) -> None:
        # Initialise the HFMetric
        super().__init__(
            tag=tag,
            config_name=config_name,
            keep_in_memory=keep_in_memory,
            seed=seed,
            experiment_id=experiment_id,
            download_config=download_config,
            **kwargs,
        )

    def compute(
        self,
        *,
        predictions: Optional[ArrayLike] = None,
        references: Optional[ArrayLike] = None,
        prediction_intervals: Optional[ArrayLike] = None,
        **kwargs: Any,
    ) -> MetricResult:
        """Computes the metric."""

        # Merge explicit keyword arguments with those stored in config
        if self._state is not None:
            kwargs = {**self._state, **kwargs}

        # Safeguard against invalid kwargs
        if not all(k in self._signature.parameters for k in kwargs):
            unknown_kwargs = [k for k in kwargs if k not in self._signature.parameters]
            raise ValueError(
                f"Unknown compute parameter(s): {unknown_kwargs}\n\n"
                f"Expected signature self.compute{self._signature}",
            )

        predictions, references = self._pre_process_inputs(
            predictions=predictions,
            references=references,
        )

        # wrap huggingface's implementation
        try:
            # This loads the data from the cache into self.data,
            # and then calls ._compute()
            result: Optional[MetricResult] = super().compute(
                predictions=predictions,
                references=references,
                prediction_intervals=prediction_intervals,
                **kwargs,
            )

            # On distributed systems, huggingface would return None
            if result is None:
                raise RuntimeError

        except Exception as exc:
            raise RuntimeError(f"Error computing metric {self.tag!r}") from exc

        return result

    def _compute(
        self,
        *,
        predictions: Optional[ArrayLike] = None,
        references: Optional[ArrayLike] = None,
        prediction_intervals: Optional[ArrayLike] = None,
        **kwargs: Any,
    ) -> MetricResult:
        """The callable invoked by .compute() to compute the metric."""

        # Do some checks common to all metrics
        if predictions is None or references is None:
            raise RuntimeError(
                f"""Could not compute metric {self.tag}: no data found in the stack.""",
            )

        return self._score(
            predictions=predictions,
            references=references,
            prediction_intervals=prediction_intervals,
            **kwargs,
        )

    @abstractmethod
    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        prediction_intervals: Optional[ArrayLike] = None,
        **kwargs: Any,
    ) -> MetricResult:
        """The callable invoked to compute the metric. To be implemented by subclasses."""



from os import PathLike as OSPathLike
from typing import Iterable, Sized, Tuple, Union

ArrayLike = Iterable
PathLike = Union[str, OSPathLike]
Splittable = Sized
SplitIndices = Tuple[Iterable[int], Iterable[int], Iterable[int]]



"""Root mean squared error regression loss."""

import logging
from importlib.metadata import version
from typing import Any, List, Optional

import evaluate
from packaging.version import parse
from sklearn.metrics import mean_squared_error

import datasets

if parse(version("scikit-learn")) >= parse("1.4.0"):
    ROOT_MEAN_SQUARED_ERROR_AVAILABLE = True
    from sklearn.metrics import root_mean_squared_error
else:
    ROOT_MEAN_SQUARED_ERROR_AVAILABLE = False


from molflux.metrics.bases import HFMetric
from molflux.metrics.typing import ArrayLike, MetricResult

logger = logging.getLogger(__name__)

_DESCRIPTION = """
The root_mean_squared_error function computes root mean square error, a risk metric
corresponding to the expected value of the (quadratic) error or loss.
"""

_KWARGS_DESCRIPTION = """
Args:
    predictions: Estimated target values.
    references: Ground truth (correct) target values.
    sample_weight (optional): Weighting of each sample.
    multioutput (optional): Defines aggregating of multiple output values.
        Array-like value defines weights used to average errors. Alternatively,
        one of {'raw_values', 'uniform_average'}. Defaults to 'uniform_average'.
        'raw_values' :
            Returns a full set of errors in case of multioutput input.
        'uniform_average' :
            Errors of all outputs are averaged with uniform weight.

Returns:
    root_mean_squared_error: A non-negative floating point value (the best value
    is 0.0), or an array of floating point values, one for each individual target.

Examples:
    >>> from molflux.metrics import load_metric
    >>> metric = load_metric("root_mean_squared_error")
    >>> predictions = [2.5, 0.0, 2, 8]
    >>> references = [3, -0.5, 2, 7]
    >>> metric.compute(predictions=predictions, references=references)
    {'root_mean_squared_error': 0.612...}
    >>> metric = load_metric("root_mean_squared_error", config_name="multioutput")
    >>> predictions = [[0, 2],[-1, 2],[8, -5]]
    >>> references = [[0.5, 1],[-1, 1],[7, -6]]
    >>> metric.compute(predictions=predictions, references=references)
    {'root_mean_squared_error': 0.822...}
"""

_CITATION = """\
@article{scikit-learn,
  title={Scikit-learn: Machine Learning in {P}ython},
  author={Pedregosa, F. and Varoquaux, G. and Gramfort, A. and Michel, V.
         and Thirion, B. and Grisel, O. and Blondel, M. and Prettenhofer, P.
         and Weiss, R. and Dubourg, V. and Vanderplas, J. and Passos, A. and
         Cournapeau, D. and Brucher, M. and Perrot, M. and Duchesnay, E.},
  journal={Journal of Machine Learning Research},
  volume={12},
  pages={2825--2830},
  year={2011}
}
"""


@evaluate.utils.file_utils.add_start_docstrings(_DESCRIPTION, _KWARGS_DESCRIPTION)
class RootMeanSquaredError(HFMetric):
    def _info(self) -> evaluate.MetricInfo:
        return evaluate.MetricInfo(
            description=_DESCRIPTION,
            citation=_CITATION,
            inputs_description=_KWARGS_DESCRIPTION,
            features=datasets.Features(
                {
                    "predictions": datasets.Sequence(datasets.Value("float")),
                    "references": datasets.Sequence(datasets.Value("float")),
                }
                if self.config_name == "multioutput"
                else {
                    "predictions": datasets.Value("float"),
                    "references": datasets.Value("float"),
                },
            ),
            reference_urls=[
                "https://scikit-learn.org/stable/modules/generated/sklearn.metrics.root_mean_squared_error.html",
            ],
        )

    def _score(
        self,
        *,
        predictions: ArrayLike,
        references: ArrayLike,
        sample_weight: Optional[List[float]] = None,
        multioutput: str = "uniform_average",
        **kwargs: Any,
    ) -> MetricResult:
        # TODO(avianello): cleanup once we can drop python 3.8 (RMSE will always be available)
        if ROOT_MEAN_SQUARED_ERROR_AVAILABLE:  # py3.9+
            score = root_mean_squared_error(
                y_true=references,
                y_pred=predictions,
                sample_weight=sample_weight,
                multioutput=multioutput,
            )
        else:  # py3.8
            score = mean_squared_error(
                y_true=references,
                y_pred=predictions,
                sample_weight=sample_weight,
                multioutput=multioutput,
                squared=False,
            )
        return {self.tag: score}



import numpy as np
import pytest
from scipy.stats import norm

from molflux.metrics import Metric, list_metrics, load_metric
from molflux.metrics.uncertainty.gaussian_nll import GaussianNLL


@pytest.fixture()
def fixture_metric():
    return load_metric("gaussian_nll")


def test_metric_in_catalogue():
    """That the metric is registered in the catalogue."""
    catalogue = list_metrics()
    all_metric_names = [name for names in catalogue.values() for name in names]
    assert "gaussian_nll" in all_metric_names


def test_metric_is_mapped_to_correct_class(fixture_metric):
    """That the catalogue name is mapped to the appropriate class."""
    metric = fixture_metric
    assert isinstance(metric, GaussianNLL)


def test_implements_protocol(fixture_metric):
    """That the metric implements the public Metric protocol."""
    metric = fixture_metric
    assert isinstance(metric, Metric)


def test_default_compute(fixture_metric):
    """That negative log likelihood computes properly."""
    metric = fixture_metric
    predictions = [0, 1, 2, 3, 4]
    references = [0.5, 0.8, 1.2, 2.4, 4.5]
    lower_bound = [0.1, 0.1, 0.2, 2, 3]
    upper_bound = [1, 1.5, 1.8, 2.8, 5.7]
    prediction_intervals = list(zip(lower_bound, upper_bound))
    result = metric.compute(
        predictions=predictions,
        references=references,
        prediction_intervals=prediction_intervals,
    )

    assert pytest.approx(result["gaussian_nll"], 0.0001) == 1.294903148251583


def test_exact_compute(fixture_metric):
    """That the negative log likelihood matches the mathematical equation."""
    metric = fixture_metric
    predictions = [0, 1, 2, 3, 4]
    references = [1, 2, 3, 4, 5]
    lower_bound = [0, 1, 2, 3, 4]
    upper_bound = [2, 3, 4, 5, 6]
    prediction_intervals = list(zip(lower_bound, upper_bound))
    confidence = 1 - 2 * (1 - norm.cdf(1))
    result = metric.compute(
        predictions=predictions,
        references=references,
        prediction_intervals=prediction_intervals,
        confidence=confidence,
    )

    assert pytest.approx(result["gaussian_nll"], 0.0001) == 1 / 2 * (
        np.log(2) + np.log(np.pi) + 1
    )


def test_raise_error_no_prediction_interval(fixture_metric):
    """That omitting prediction interval raises error."""
    metric = fixture_metric
    predictions = [3, 0.0, 2]
    references = [3, -0.0, 1]
    # prediction_intervals bound argument omitted
    with pytest.raises(RuntimeError):
        metric.compute(predictions=predictions, references=references)



try:
    import torch
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("lightning", e) from e

# NOTE Some schedulers disabled due to complex initialisation
# Would require a bit more custom logic (but could be done)


AVAILABLE_SCHEDULERS = {
    # "LambdaLR": torch.optim.lr_scheduler.LambdaLR,
    "StepLR": torch.optim.lr_scheduler.StepLR,
    "MultiStepLR": torch.optim.lr_scheduler.MultiStepLR,
    "ConstantLR": torch.optim.lr_scheduler.ConstantLR,
    "LinearLR": torch.optim.lr_scheduler.LinearLR,
    "ExponentialLR": torch.optim.lr_scheduler.ExponentialLR,
    "CosineAnnealingLR": torch.optim.lr_scheduler.CosineAnnealingLR,
    # "ChainedScheduler": torch.optim.lr_scheduler.ChainedScheduler,
    # "SequentialLR": torch.optim.lr_scheduler.SequentialLR,
    "ReduceLROnPlateau": torch.optim.lr_scheduler.ReduceLROnPlateau,
    "CyclicLR": torch.optim.lr_scheduler.CyclicLR,
    "CosineAnnealingWarmRestarts": torch.optim.lr_scheduler.CosineAnnealingWarmRestarts,
    "OneCycleLR": torch.optim.lr_scheduler.OneCycleLR,
}



from os import PathLike as OSPathLike
from typing import (
    Any,
    Dict,
    List,
    Protocol,
    Union,
    runtime_checkable,
)

import datasets


@runtime_checkable
class SupportsDataframeInterchangeProtocol(Protocol):
    """Protocol representing any object conforming to the Dataframe Interchange Protocol.

    Several DataFrame libraries now support this interchange protocol (pandas, modin, vaex, cudf, pyarrow).

    References:
        https://data-apis.org/dataframe-protocol/latest/purpose_and_scope.html
        https://data-apis.org/dataframe-protocol/latest/API.html
    """

    def __dataframe__(self, nan_as_null: bool = False, allow_copy: bool = True) -> Any:
        ...


DataFrameLike = Union[SupportsDataframeInterchangeProtocol, datasets.Dataset]
Classes = Dict[str, List[Any]]
Features = List[str]
PathLike = Union[str, OSPathLike]
PredictionResult = Dict[str, List[Any]]



import pandas as pd
import pytest

import datasets
from molflux.modelzoo.load import load_model
from molflux.modelzoo.protocols import Estimator


def test_load():
    model = load_model("mock_model")
    assert model is not None


def test_implements_protocol():
    model = load_model("mock_model")
    assert isinstance(model, Estimator)


train_df = pd.DataFrame({"x1": ["Hello"], "y1": ["Buongiorno"]})


@pytest.mark.parametrize(
    "train_data",
    [
        train_df,
        datasets.Dataset.from_pandas(train_df),
    ],
)
def test_train(train_data):
    model = load_model("mock_model", x_features=["x1"], y_features=["y1"])
    model.train(train_data=train_data)
    assert True



from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    Optional,
    Protocol,
    Sized,
    Tuple,
    runtime_checkable,
)


@runtime_checkable
class Representation(Protocol):
    def featurise(self, samples: Iterable[Any], **kwargs: Any) -> Dict[str, Any]:
        """Featurises the input samples."""


@runtime_checkable
class Representations(Protocol):
    def __iter__(self) -> Iterator[Representation]:
        ...


@runtime_checkable
class SplittingStrategy(Protocol):
    def split(
        self,
        dataset: Sized,
        y: Optional[Iterable] = None,
        groups: Optional[Iterable] = None,
        **kwargs: Any,
    ) -> Iterator[Tuple[Iterable[int], Iterable[int], Iterable[int]]]:
        """Generates indices to split data into training, validation, and test sets."""



import logging
from dataclasses import asdict, field
from typing import Any, Callable, Dict, List, Literal, Optional, TypeVar, Union

from pydantic import validator
from pydantic.dataclasses import dataclass

from molflux.modelzoo.model import ModelConfig
from molflux.modelzoo.models.lightning.trainer.callbacks.stock_callbacks import (
    AVAILABLE_CALLBACKS,
)
from molflux.modelzoo.models.lightning.trainer.schedulers.stock_schedulers import (
    AVAILABLE_SCHEDULERS,
)

try:
    from lightning.pytorch import Trainer
    from lightning.pytorch import callbacks as pl_callbacks
    from lightning.pytorch import loggers as pl_loggers
    from lightning.pytorch import profilers as pl_profilers
    from lightning.pytorch import strategies as pl_strategies
    from torch.optim import Optimizer
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("lightning", e) from e


def make_dvc_logger(**kwargs: Any) -> Any:
    try:
        from dvclive.lightning import DVCLiveLogger  # pyright: ignore

        logging.getLogger("dvc_studio_client.post_live_metrics").setLevel(
            logging.WARNING,
        )

    except ImportError as err:
        from exs.modelzoo.errors import OptionalDependencyImportError

        raise OptionalDependencyImportError("DVCLive", "dvclive") from err

    return DVCLiveLogger(**kwargs)


AVAILABLE_LOGGERS: Dict[str, Callable] = {
    "CometLogger": pl_loggers.CometLogger,
    "CSVLogger": pl_loggers.CSVLogger,
    "DVCLiveLogger": make_dvc_logger,
    "MLFlowLogger": pl_loggers.MLFlowLogger,
    "NeptuneLogger": pl_loggers.NeptuneLogger,
    "TensorBoardLogger": pl_loggers.TensorBoardLogger,
    "WandbLogger": pl_loggers.WandbLogger,
}

AVAILABLE_PROFILERS = {
    "SimpleProfiler": pl_profilers.SimpleProfiler,
    "AdvancedProfiler": pl_profilers.AdvancedProfiler,
    "PyTorchProfiler": pl_profilers.PyTorchProfiler,
}


AVAILABLE_STRATEGIES = {
    "FSDPStrategy": pl_strategies.FSDPStrategy,
    "DDPStrategy": pl_strategies.DDPStrategy,
    "DeepSpeedStrategy": pl_strategies.DeepSpeedStrategy,
}


class ConfigDict:
    extra = "forbid"
    arbitrary_types_allowed = True
    smart_union = True


def _dict_is_single_logger_config(logger_dict: Dict[str, Any]) -> bool:
    return (
        ("name" in logger_dict)
        and ("config" in logger_dict)
        and (isinstance(logger_dict["name"], str))
    )


@dataclass(config=ConfigDict)
class TrainerConfig:
    """
    Lightning Trainer config.
    See https://lightning.ai/docs/pytorch/stable/common/trainer.html#trainer-class-api for more info.
    """

    accelerator: str = "auto"
    accumulate_grad_batches: int = 1
    barebones: bool = False
    benchmark: Optional[bool] = None
    callbacks: Union[List[Dict[str, Any]], Dict[str, Dict[str, Any]], None] = None
    check_val_every_n_epoch: Optional[int] = 1
    default_root_dir: Optional[str] = "training"
    detect_anomaly: bool = False
    deterministic: Union[bool, Literal["warn"], None] = None
    devices: Union[List[int], str, int] = "auto"
    enable_checkpointing: bool = False  # NOTE vanilla PL default is True
    enable_model_summary: bool = True
    enable_progress_bar: bool = True
    fast_dev_run: Union[int, bool] = False
    gradient_clip_algorithm: Optional[str] = None
    gradient_clip_val: Optional[Union[int, float]] = None
    inference_mode: bool = True
    limit_predict_batches: Union[int, float, None] = None
    limit_test_batches: Union[int, float, None] = None
    limit_train_batches: Union[int, float, None] = None
    limit_val_batches: Union[int, float, None] = None
    log_every_n_steps: int = 50
    logger: Union[
        Dict[str, Any],
        List[Dict[str, Any]],
        Dict[str, Dict[str, Any]],
        bool,
        None,
    ] = True
    max_epochs: Optional[int] = 1
    max_steps: int = -1
    max_time: Union[str, Dict[str, int], None] = None
    min_epochs: Optional[int] = None
    min_steps: Optional[int] = None
    num_nodes: int = 1
    num_sanity_val_steps: Optional[int] = None
    overfit_batches: Union[int, float] = 0.0
    precision: Union[int, str] = 32
    profiler: Optional[Dict[str, Any]] = None
    reload_dataloaders_every_n_epochs: int = 0
    strategy: Union[str, Dict[str, Any]] = "auto"
    sync_batchnorm: bool = False
    use_distributed_sampler: bool = True
    val_check_interval: Union[int, float, None] = None

    def pass_to_trainer(self) -> Dict[str, Any]:
        out = asdict(self)
        out["callbacks"] = self.convert_callbacks(out["callbacks"])
        out["logger"] = self.convert_logger(out["logger"])
        out["profiler"] = self.convert_profiler(out["profiler"])
        out["strategy"] = self.convert_strategy(out["strategy"])

        return out

    def convert_callbacks(
        self,
        callbacks: Union[List[Dict[str, Any]], Dict[str, Dict[str, Any]], None],
    ) -> Optional[List[pl_callbacks.Callback]]:
        if callbacks is None:
            return callbacks

        if isinstance(callbacks, dict):
            callbacks = list(callbacks.values())

        return [
            AVAILABLE_CALLBACKS[callback["name"]](**callback.get("config", {}))
            for callback in callbacks
        ]

    def convert_logger(
        self,
        logger: Union[
            Dict[str, Any],
            List[Dict[str, Any]],
            Dict[str, Dict[str, Any]],
            bool,
            None,
        ],
    ) -> Union[List[pl_loggers.Logger], bool, None]:
        if logger is None or isinstance(logger, bool):
            return logger

        if isinstance(logger, dict):
            if _dict_is_single_logger_config(logger):
                logger = [logger]
            else:
                logger = list(logger.values())

        return [
            AVAILABLE_LOGGERS[logger_["name"]](**logger_.get("config", {}))
            for logger_ in logger
        ]

    def convert_profiler(
        self,
        profiler: Optional[Dict[str, Any]],
    ) -> Optional[pl_profilers.Profiler]:
        if profiler is None:
            return profiler

        out: pl_profilers.Profiler = AVAILABLE_PROFILERS[profiler["name"]](
            **profiler.get("config", {}),
        )
        return out

    def convert_strategy(
        self,
        strategy: Union[Dict[str, Any], str],
    ) -> Union[pl_strategies.Strategy, str]:
        if isinstance(strategy, str):
            return strategy

        out: pl_strategies.Strategy = AVAILABLE_STRATEGIES[strategy["name"]](
            **strategy.get("config", {}),
        )
        return out


@dataclass(config=ConfigDict)
class OptimizerConfig:
    name: str = "Adam"
    config: Dict[str, Any] = field(default_factory=lambda: {"lr": 1e-4})


@dataclass(config=ConfigDict)
class SchedulerConfig:
    name: str = "CosineAnnealingLR"
    config: Dict[str, Any] = field(default_factory=lambda: {"T_max": "num_steps"})
    interval: Literal["epoch", "step"] = "step"
    frequency: int = 1
    monitor: str = "val_loss"
    strict: bool = True

    def prepare_scheduler(
        self,
        optimizer: Optimizer,
        trainer: Trainer,
    ) -> Dict[str, Any]:
        # NOTE Enables config-driven decay of learning rate to final step
        if self.config.get("T_max", None) == "num_steps":
            self.config["T_max"] = trainer.estimated_stepping_batches
            self.interval = "step"
        elif self.config.get("T_max", None) == "num_epochs":
            self.config["T_max"] = trainer.max_epochs
            self.interval = "epoch"

        if self.config.get("total_steps", None) == "num_steps":
            self.config["total_steps"] = trainer.estimated_stepping_batches
            self.interval = "step"
        elif self.config.get("total_steps", None) == "num_epochs":
            self.config["total_steps"] = trainer.max_epochs
            self.interval = "epoch"

        return {
            "scheduler": AVAILABLE_SCHEDULERS[self.name](optimizer, **self.config),
            "interval": self.interval,
            "frequency": self.frequency,
            "monitor": self.monitor,
            "strict": self.strict,
        }


@dataclass(config=ConfigDict)
class TransferLearningStage:
    freeze_modules: Optional[List[str]] = None
    trainer: Optional[Dict[str, Any]] = None
    datamodule: Optional[Dict[str, Any]] = None
    optimizer: Optional[Dict[str, Any]] = None
    scheduler: Optional[Dict[str, Any]] = None

    @validator("trainer")
    def must_fit_into_trainer(
        cls,
        trainer: Optional[Dict[str, Any]],
    ) -> Optional[Dict[str, Any]]:
        if trainer is not None:
            TrainerConfig(**trainer)
        return trainer

    @validator("datamodule")
    def must_fit_into_datamodule(
        cls,
        datamodule: Optional[Dict[str, Any]],
    ) -> Optional[Dict[str, Any]]:
        if datamodule is not None:
            DataModuleConfig(**datamodule)
        return datamodule

    @validator("optimizer")
    def must_fit_into_optimizer(
        cls,
        optimizer: Optional[Dict[str, Any]],
    ) -> Optional[Dict[str, Any]]:
        if optimizer is not None:
            OptimizerConfig(**optimizer)
        return optimizer

    @validator("scheduler")
    def must_fit_into_scheduler(
        cls,
        scheduler: Optional[Dict[str, Any]],
    ) -> Optional[Dict[str, Any]]:
        if scheduler is not None:
            SchedulerConfig(**scheduler)
        return scheduler


@dataclass(config=ConfigDict)
class TransferLearningConfigBase:
    stages: List[TransferLearningStage]
    pre_trained_model_path: Optional[str] = None
    repo_url: Optional[str] = None
    rev: Optional[str] = None
    model_path_in_repo: str = "model"
    modules_to_match: Optional[Dict[str, str]] = None

    def __post_init__(self) -> None:
        if (self.pre_trained_model_path is not None) ^ (
            (self.repo_url is not None)
            and (self.rev is not None)
            and (self.model_path_in_repo is not None)
        ):
            pass
        else:
            raise ValueError(
                "Must specify either 'pre_trained_model_path' or all of 'repo_url', 'rev', 'model_path_in_repo' (but not both).",
            )

        for i in range(len(self.stages)):
            if isinstance(self.stages[i], dict):
                self.stages[i] = TransferLearningStage(**self.stages[i])  # type: ignore


@dataclass(config=ConfigDict)
class SplitConfig:
    """batch_size: The batch size to use for this data split, optionally as a
    dictionary assigning different batch sizes to different datasets."""

    batch_size: Union[int, Dict[str, int]] = 1


@dataclass(config=ConfigDict)
class TrainSplitConfig(SplitConfig):
    drop_last: bool = True
    mode: Literal[
        "max_size_cycle",
        "max_size",
        "min_size",
    ] = "max_size_cycle"


@dataclass(config=ConfigDict)
class EvalSplitConfig(SplitConfig):
    mode: Literal[
        "max_size_cycle",
        "max_size",
        "min_size",
        "sequential",
    ] = "max_size"


@dataclass(config=ConfigDict)
class DataModuleConfig:
    train: TrainSplitConfig = field(default_factory=TrainSplitConfig)
    validation: EvalSplitConfig = field(default_factory=EvalSplitConfig)
    test: EvalSplitConfig = field(default_factory=EvalSplitConfig)
    predict: SplitConfig = field(default_factory=SplitConfig)
    num_workers: Union[int, Literal["all"]] = 0


@dataclass(config=ConfigDict)
class CompileConfig:
    mode: Literal["default", "reduce-overhead", "max-autotune"]
    dynamic: bool = False
    fullgraph: bool = False
    backend: str = "inductor"
    backend_kwargs: Dict[str, Any] = field(default_factory=dict)


@dataclass(config=ConfigDict)
class LightningConfig(ModelConfig):
    datamodule: DataModuleConfig = field(default_factory=DataModuleConfig)
    trainer: TrainerConfig = field(default_factory=TrainerConfig)
    optimizer: OptimizerConfig = field(default_factory=OptimizerConfig)
    scheduler: Optional[SchedulerConfig] = None
    transfer_learning: Optional[TransferLearningConfigBase] = None
    compile: Union[CompileConfig, bool] = False
    # NOTE This setting affects a global variable!
    float32_matmul_precision: Optional[Literal["highest", "high", "medium"]] = None

    @validator("compile", pre=True)
    def compile_true_to_compile_config_with_default_mode(
        cls,
        compile: Union[Dict[str, Any], bool],
    ) -> Union[Dict[str, Any], bool]:
        if compile is True:
            return {"mode": "default"}
        return compile


LightningConfigT = TypeVar("LightningConfigT", bound=LightningConfig)



from os import PathLike as OSPathLike
from typing import (
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import fsspec

from datasets import Dataset, DatasetDict

# A list of lists of strings or Nones
DisplayNames = List[List[Optional[str]]]

FileSystem = fsspec.AbstractFileSystem
ExamplesGenerator = Iterator[Tuple[Union[int, str], Dict]]

# This includes str, pathlib.Path, cloudpathlib.CloudPath, ...
PathLike = Union[str, OSPathLike]

# Huggingface's DataFiles
HFDataFiles = Union[str, Sequence[str], Mapping[str, Union[str, Sequence[str]]]]
# An enhanced version accepting PathLike objects too
DataFiles = Union[
    PathLike,
    Sequence[PathLike],
    Mapping[str, Union[PathLike, Sequence[PathLike]]],
]

DatasetType = TypeVar("DatasetType", Dataset, DatasetDict)



import json
from collections import defaultdict
from enum import Enum
from typing import Any, Dict, List, Optional, Sequence, Type, Union

from more_itertools import always_iterable

from molflux.features.bases import RepresentationBase, RepresentationInfo
from molflux.features.representations.openeye._utils import (
    get_sd_data,
    iterate_mols_or_confs,
    to_oemol,
    to_smiles,
)
from molflux.features.typing import ArrayLike
from molflux.features.utils import featurisation_error_harness


class DType(Enum):
    """DType callable Enum. Used to stringify data types that featurisations
    can be cast to when parsing SD tags."""

    FLOAT: Type = float
    INT: Type = int
    STR: Type = str

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Call method to call the underlying callable's method."""
        return self.value(*args, **kwargs)


class SDFeature(RepresentationBase):
    """Representation used to retrieve precalculated featurisations from SD tags
    on molecules. Featurisations are retrieved as if they are JSON strings unless
    otherwise specified."""

    def _info(self) -> RepresentationInfo:
        """Representation Info for SDFeature type Representation."""
        return RepresentationInfo(
            description=(
                "Prism Representation that retrieves a featurisation from "
                "an SD tag(s) by name."
            ),
        )

    def _featurise(
        self,
        samples: ArrayLike,
        *,
        sd_tags: Optional[Union[str, List[str]]] = None,
        as_json: Union[bool, List[str], Dict[str, bool]] = True,
        dtype: Union[DType, str, Dict[str, Union[str, DType]]] = "str",
        separator: Union[str, Dict[str, str]] = "\n",
        **kwargs: Any,
    ) -> Dict[str, List[Any]]:
        """Featurise method for SDFeature Representation. Defines the SD tags to
        check for precalculated representations and defines how to parse each SD
        tag.

        Args:
            samples: Samples that must be OEMolBase parsable.
            sd_tags: The SD tags to retrieve the featurisations of.
            as_json: Whether to parse all SD tags as JSON tags (`True`), or names of
                SD tags in `sd_tags` to parse as JSON tags, or a mapping of
                which tags to parse as JSON values and which to not. Tags in `sd_tags`
                that are not parsed as JSON are expected to be `separator`
                separated strings and are parsed as such.
            dtype: The data-type to parse each split molflux of an SD tag representation.
                Either a single type to apply to all, or a mapping between SD tags
                and the dtype to parse them as. Only applicable to SD tags that are
                *not* parsed `as_json`. Default is `"str"`.
            separator: The separator to split each feature molflux of an SD tag
                representation. Either a single split to apply to all `sd_tags`,
                or a mapping between SD tags and the `separator` to split
                molfluxs with. Only applicable to SD tags that are *not* parsed
                `as_json`. Default is `"\\n"`.

        Returns:
            A dictionary with the `sd_tags` as keys and a list of length
            `len(samples)` with parsed featurisations as values.

        Raises:
            ValueError: If the required `sd_tags` are not present on a molecule.

        Notes:
            If passing multi conformer molecules, its assumed that the SD tags
            will be present on the conformer, however if not present it this
            representation will fall back onto the parent molecules SD tags.

        Examples:
            >>> from molflux.features import load_representation
            >>> from openeye import oechem
            >>> import json

            >>> representation = load_representation("sd_feature")

            >>> mol = oechem.OEGraphMol()
            >>> assert oechem.OESmilesToMol(mol, "CCC")
            >>> assert oechem.OESetSDData(mol, "featurisation", json.dumps([0, 1, 1, 1, 0]))
            >>> assert oechem.OESetSDData(mol, "separated_features", "feature1\\nfeature2\\nfeature3")
            >>> # Ask for these specific tags.
            >>> representation.update_state(sd_tags=["featurisation", "separated_features"], as_json=["featurisation"])
            >>> representation.featurise([mol])
            {'sd_feature::featurisation': [[0, 1, 1, 1, 0]], 'sd_feature::separated_features': [['feature1', 'feature2', 'feature3']]}
        """
        if not sd_tags:
            raise ValueError(
                "Must supply SD tags to retrieve featurisations for. Supplied:"
                f" {sd_tags}",
            )
        sd_tags_list: Sequence[str] = list(always_iterable(sd_tags))

        # If a sequence is supplied, default should be False, otherwise True.
        # This is because if someone supplies explicit tags to parse as JSON,
        # it would be expected that the others are *not* parsed as JSON
        default_as_json = not isinstance(as_json, Sequence)

        as_json_mapping = {}
        sep_mapping = {}
        dtype_mapping = {}

        # Populate the above mappings - this happens here to avoid having to
        # determine each for each sd tag *within the inner loop over molecules*
        # (i.e. determining the parsing for each SD tag for each mol).
        for tag in sd_tags_list:
            # Explicitly supplied mapping of which tags to parse as JSON (and not)
            if isinstance(as_json, dict):
                # Don't populate with tags not supplied.
                if tag in as_json:
                    as_json_mapping[tag] = as_json[tag]
            elif isinstance(as_json, bool):
                # Parse all tags or not
                as_json_mapping[tag] = as_json
            elif isinstance(as_json, Sequence):
                # List of tags to parse as JSON
                as_json_mapping[tag] = tag in as_json

            # Only populate these if they are *not* to be parsed as JSON
            if not as_json_mapping.get(tag, default_as_json):
                if isinstance(separator, dict):
                    # Don't populate with tags not supplied.
                    if tag in separator:
                        sep_mapping[tag] = separator[tag]
                elif isinstance(separator, str):
                    # Parse all tags or not
                    sep_mapping[tag] = separator

                if isinstance(dtype, dict):
                    # Don't populate with tags not supplied.
                    if tag in dtype:
                        dt = dtype[tag]
                        dtype_mapping[tag] = (
                            DType[dt.upper()] if not isinstance(dt, DType) else dt
                        )
                elif isinstance(dtype, (str, DType)):
                    # Parse all tags or not
                    dtype_mapping[tag] = (
                        DType[dtype.upper()] if not isinstance(dtype, DType) else dtype
                    )

        # To hold the feature representations.
        features = defaultdict(list)

        # Turn to oemolecules
        mols = map(to_oemol, samples)
        for mol in iterate_mols_or_confs(mols):
            with featurisation_error_harness(to_smiles(mol)):
                for tag in sd_tags_list:
                    parse_json = as_json_mapping.get(tag, default_as_json)
                    # This will raise if the tag is not present.
                    value = get_sd_data(mol, tag=tag, dtype=str)

                    if parse_json:
                        value = json.loads(value)
                    else:
                        # Split the value by the separator and cast as datatype.
                        dtype_type = dtype_mapping.get(tag, DType["STR"])
                        tag_sep = sep_mapping.get(tag, "\n")
                        value = [*map(dtype_type, value.split(tag_sep))]
                    features[f"{self.tag}::{tag}"].append(value)

        return dict(features)



from os import PathLike as OSPathLike
from typing import Any, Dict, Iterable, List, Union

RepresentationResult = Dict[str, Any]
ArrayLike = Iterable[Any]
SmilesArray = Iterable[str]
OEMolArray = Iterable[Any]
MolArray = Union[SmilesArray, OEMolArray, Iterable[bytes]]

PathLike = Union[str, OSPathLike]
Fingerprint = List[int]



import logging
from typing import Any

try:
    import torch
    from lightning.pytorch import LightningModule, Trainer, callbacks
except ImportError as e:
    from molflux.modelzoo.errors import ExtrasDependencyImportError

    raise ExtrasDependencyImportError("lightning", e) from e


logger = logging.getLogger(__name__)


class ModelCheckpointApply(callbacks.ModelCheckpoint):
    """Model checkpointing callback which applies the best module checkpoint at the end of fitting"""

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    def on_fit_end(self, trainer: Trainer, pl_module: LightningModule) -> None:
        if self.best_model_path is not None:
            best_ckpt = torch.load(self.best_model_path, map_location="cpu")
            pl_module.load_state_dict(best_ckpt["state_dict"])
            logger.warning(
                f"Using best model from ckpt_path: {self.best_model_path}.\n",
            )


AVAILABLE_CALLBACKS = {
    "BackboneFinetuning": callbacks.BackboneFinetuning,
    "BaseFinetuning": callbacks.BaseFinetuning,
    "BasePredictionWriter": callbacks.BasePredictionWriter,
    "DeviceStatsMonitor": callbacks.DeviceStatsMonitor,
    "EarlyStopping": callbacks.EarlyStopping,
    "GradientAccumulationScheduler": callbacks.GradientAccumulationScheduler,
    "LearningRateMonitor": callbacks.LearningRateMonitor,
    "ModelCheckpoint": callbacks.ModelCheckpoint,
    "ModelCheckpointApply": ModelCheckpointApply,
    "ModelPruning": callbacks.ModelPruning,
    "ModelSummary": callbacks.ModelSummary,
    "RichModelSummary": callbacks.RichModelSummary,
    "StochasticWeightAveraging": callbacks.StochasticWeightAveraging,
    "Timer": callbacks.Timer,
}
